/**
   This program evaluates (interprets) an Abstract Syntax Tree (AST)
   that is defined by the following grammar.

       Prog ::= Exp
              | '(' 'prog' Exp+ ')'

        Exp ::= If
              | While
              | Var
              | Set
              | Print
              | BExp
              | AExp
              | INTEGER
              | BOOLEAN
              | VARIABLE

         If ::= '(' 'if' Exp Exp Exp ')'

      While ::= '(' 'while' Exp Exp+ ')'

        Var ::= '(' 'var' VARIABLE Exp ')'

        Set ::= '(' 'var' VARIABLE Exp ')'

      Print ::= '(' 'print' Exp ')'

       BExp ::= '(' '||'  Exp Exp+ ')'
              | '(' '&&'  Exp Exp+ ')'
              | '(' '!'   Exp ')'
              | '(' RelOp Exp Exp ')'
              | '('  EqOp Exp Exp ')'

      RelOp ::= '<' | '>' | '<=' | '>='
       EqOp ::= '==' | '!='

       AExp ::= '(' '+' Exp Exp* ')'
              | '(' '-' Exp Exp? ')'
              | '(' '*' Exp Exp+ ')'
              | '(' '/' Exp Exp  ')'
              | '(' '%' Exp Exp  ')'
              | '(' '^' Exp Exp  ')'

    INTEGER ::= [-][0-9]+
    BOOLEAN ::= 'true' | 'false'
   VARIABLE ::= [a-zA-Z][a-zA-Z0-9]*
*/

public class Evaluate
{
   private static final int DEBUG = 1;

   private static Environment env;  // global environment data structure


   /**
      The methods evaluateProg(), evaluateIf(), evaluateWhile(),
      evaluateVar(), evaluateSet(), evaluatePrint(), evaluateExp(),
      evaluateAExp(), evaluateBExp(), and evaluateRExp() are
      essentially a post-order traversal of the abstract syntax tree.
   */
   public static Value eval(Tree tree) throws EvalException
   {
      return evaluateProg( tree );
   }//eval()


   // Evaluate a prog expression.
   private static Value evaluateProg(Tree tree) throws EvalException
   {
      Value result = null;

      // Instantiate the global environment.
      env = new Environment();

      // Check which kind of Prog we have.
      if ( ! tree.getElement().equals("prog") )
      {
         // Evaluate the single expression.
         result = evaluateExp( tree );
      }
      else
      {
         // Evaluate each Exp in the Prog.
         // Any Var expressions will have the side effect
         // of putting a variable in the global environment.
         // Any Set expressions will have the side effect
         // of changing a value in the global environment.
         // Any Print expressions will have the side effect
         // of printing an output.
         for (int i = 0; i < tree.degree() - 1; i++)
         {
            evaluateExp( tree.getSubTree(i) );
         }

         // Evaluate the last expression and use its
         // value as the value of the prog expression.
         result = evaluateExp( tree.getSubTree(tree.degree() - 1) );
      }

      return result;
   }//evaluateProg()


   // Evaluate an expression.
   private static Value evaluateExp(Tree tree) throws EvalException
   {
      Value result = null;

      String node = tree.getElement();

      if ( node.equals("if") )
         result = evaluateIf( tree );
      else if ( node.equals("while") )
         result = evaluateWhile( tree );
      else if ( node.equals("var") )
         result = evaluateVar( tree );
      else if ( node.equals("set") )
         result = evaluateSet( tree );
      else if ( node.equals("print") )
         result = evaluatePrint( tree );
      else if ( node.equals("&&") || node.equals("||")
             || node.equals("!") )
         result = evaluateBExp(tree);  // boolean expression
      else if ( node.equals("<")  || node.equals(">")
             || node.equals("<=") || node.equals(">=")
             || node.equals("==") || node.equals("!=") )
         result = evaluateRExp(tree);  // relational operator
      else if ( node.equals("+") || node.equals("-")
             || node.equals("*") || node.equals("/")
             || node.equals("%") || node.equals("^") )
         result = evaluateAExp(tree);  // arithmetic expression
      else if ( tree.degree() == 0 )
      {
         if ( node.equals("true") || node.equals("false") )
            result = new Value( node.equals("true") );
         else if ( node.matches("^[-]*[0-9][0-9]*") )
            result = new Value( Integer.parseInt( node ) );
         else if ( env.defined(node) )  // a variable
            result = env.lookUp( node );
         else  // runtime check
            throw new EvalException("undefined variable: " + node + "\n");
      }
      else
         throw new EvalException("invalid expression: " + tree + "\n");

      return result;
   }//evaluateExp()


   // Evaluate an if-expression.
   private static Value evaluateIf(Tree tree) throws EvalException
   {
      if ( 3 != tree.degree() )  // runtime check
      {
         throw new EvalException("incorrect conditional expression: " + tree + "\n");
      }

      Value result = null;

      // Evaluate the boolean expression.
      Value conditionalExp = evaluateExp( tree.getSubTree(0) );
      // Do a runtime type check to make sure we have a boolean value.
      if ( ! conditionalExp.tag.equals(Value.BOOL_TAG) )
      {
         throw new EvalException("illegal boolean expression: " + tree);
      }

      // Use the boolean value to choose a branch of the if-expression.
      if ( conditionalExp.valueB )
      {
         result = evaluateExp( tree.getSubTree(1) );
      }
      else
      {
         result = evaluateExp( tree.getSubTree(2) );
      }

      return result;
   }//evaluateIf()


   // Evaluate a while-loop expression.
   private static Value evaluateWhile(Tree tree) throws EvalException
   {
      if ( tree.degree() < 2 )  // runtime check
      {
         throw new EvalException("incorrect while expression: " + tree + "\n");
      }

      Value result = null;

      // Evaluate the boolean expression.
      Value conditionalExp = evaluateExp( tree.getSubTree(0) );
      // Do a runtime type check to make sure we have a boolean value.
      if ( ! conditionalExp.tag.equals(Value.BOOL_TAG) )
      {
         throw new EvalException("illegal boolean expression: " + tree);
      }

      // Use the boolean value to control the looping.
      while ( conditionalExp.valueB )
      {
         // Evaluate each Exp in the body of the loop (for its side effect).
         for (int i = 1; i < tree.degree(); i++)
         {
            evaluateExp( tree.getSubTree(i) );
         }

         // Re-evaluate the boolean expression.
         conditionalExp = evaluateExp( tree.getSubTree(0) );
         // Do a runtime type check to make sure we have a boolean value.
         if ( ! conditionalExp.tag.equals(Value.BOOL_TAG) )
         {
            throw new EvalException("illegal boolean expression: " + tree);
         }
      }

      // Always return false for a while-loop expression.
      result =  new Value( false );

      return result;
   }//evaluateWhile()


   // Evaluate a var expression.
   private static Value evaluateVar(Tree tree) throws EvalException
   {
      if ( 2 != tree.degree() )  // runtime check
      {
         throw new EvalException("wrong number of arguments: " + tree + "\n");
      }

      Value result = null;

      // Get the variable name.
      String variable = tree.getSubTree(0).getElement();

      // Check if this variable has already been
      // declared in the environment.
      if ( env.defined(variable) )  // runtime check
      {
         throw new EvalException("variable already declared: " + variable + "\n");
      }

      // Get, and then evaluate, the expression.
      Tree expr = tree.getSubTree(1);
      result = evaluateExp( expr );

      // Create the new variable.
      env.add(variable, result);

      if (DEBUG > 0) System.out.println( env + "\n" ); // for debugging purposes

      return result;
   }//evaluateVar()


   // Evaluate a set expression.
   private static Value evaluateSet(Tree tree) throws EvalException
   {
      if ( 2 != tree.degree() )  // runtime check
      {
         throw new EvalException("wrong number of arguments: " + tree + "\n");
      }

      Value result = null;

      // Get the variable name.
      String variable = tree.getSubTree(0).getElement();

      // Check if this variable has already been
      // declared in the environment.
      if ( ! env.defined(variable) )  // runtime check
      {
         throw new EvalException("undefined variable: " + variable);
      }

      // Get, and then evaluate, the expression.
      Tree expr = tree.getSubTree(1);
      result = evaluateExp(expr);

      // Update this variable in the environment.
      env.update(variable, result);

      if (DEBUG > 0) System.out.println( env + "\n" ); // for debugging purposes

      return result;
   }//evaluateSet()


   // Evaluate a print expression.
   private static Value evaluatePrint(Tree tree) throws EvalException
   {
      if ( 1 != tree.degree() )  // runtime check
      {
         throw new EvalException("wrong number of arguments: " + tree + "\n");
      }

      Value result = evaluateExp( tree.getSubTree(0) );

      System.out.print( result + "\n" );

      return result;
   }//evaluatePrint()


   // Evaluate a boolean expression.
   private static Value evaluateBExp(Tree tree) throws EvalException
   {
      boolean result = false;

      String node = tree.getElement();

      Value value = evaluateExp( tree.getSubTree(0) );
      if ( ! value.tag.equals(Value.BOOL_TAG) )  // runtime check
      {
         throw new EvalException("not a boolean expression: "
                                  + tree.getSubTree(0) + "\n");
      }
      result = value.valueB;

      if ( node.equals("&&") )
      {
         if ( 2 > tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }

         for (int i = 1; i < tree.degree(); i++)
         {
            if (result)
            {
               value = evaluateExp( tree.getSubTree(i) );
               if ( ! value.tag.equals(Value.BOOL_TAG) )  // runtime check
               {
                  throw new EvalException("not a boolean expression: "
                                           + tree.getSubTree(i) + "\n");
               }
               result = result && value.valueB;
            }
            else  // short circuit the evaluation of '&&'
            {
               result = false;
               break;
            }
         }
      }
      else if ( node.equals("||") )
      {
         if ( 2 > tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }

         for (int i = 1; i < tree.degree(); i++)
         {
            if (! result)
            {
               value = evaluateExp( tree.getSubTree(i) );
               if ( ! value.tag.equals(Value.BOOL_TAG) )  // runtime check
               {
                  throw new EvalException("not a boolean expression: "
                                           + tree.getSubTree(i) + "\n");
               }
               result = result || value.valueB;
            }
            else  // short circuit the evaluation of '||'
            {
               result = true;
               break;
            }
         }
      }
      else if ( node.equals("!") )
      {
         if ( 1 != tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }
         result = ! result;
      }

      return new Value( result );
   }//evaluateBExp()


   // Evaluate a relational expression (which is a kind of boolean expression).
   private static Value evaluateRExp(Tree tree) throws EvalException
   {
      if ( 2 != tree.degree() )  // runtime check
      {
         throw new EvalException("wrong number of arguments: " + tree + "\n");
      }

      boolean result = false;

      String op = tree.getElement();

      Value valueL = evaluateExp( tree.getSubTree(0) );
      if ( ! valueL.tag.equals(Value.INT_TAG) )  // runtime check
      {
         throw new EvalException("not a integer expression: "
                                  + tree.getSubTree(0) + "\n");
      }

      Value valueR = evaluateExp( tree.getSubTree(1) );
      if ( ! valueR.tag.equals(Value.INT_TAG) )  // runtime check
      {
         throw new EvalException("not a integer expression: "
                                  + tree.getSubTree(1) + "\n");
      }

      int resultL = valueL.valueI;
      int resultR = valueR.valueI;

      if ( op.equals("<") )
      {
         result = resultL < resultR;
      }
      else if ( op.equals(">") )
      {
         result = resultL > resultR;
      }
      else if ( op.equals("<=") )
      {
         result = resultL <= resultR;
      }
      else if ( op.equals(">=") )
      {
         result = resultL >= resultR;
      }
      else if ( op.equals("==") )
      {
         result = resultL == resultR;
      }
      else if ( op.equals("!=") )
      {
         result = resultL != resultR;
      }

      return new Value( result );
   }//evaluateRExp()


   // Evaluate an arithmetic expression.
   private static Value evaluateAExp(Tree tree) throws EvalException
   {
      int result = 0;

      String node = tree.getElement();

      Value valueL = evaluateExp( tree.getSubTree(0) );
      if ( ! valueL.tag.equals(Value.INT_TAG) )  // runtime check
      {
         throw new EvalException("not a integer expression: "
                                  + tree.getSubTree(0) + "\n");
      }
      int resultL = valueL.valueI;
      int resultR = 0;

      Value valueR = null;
      if ( tree.degree() >= 2 )
      {
         valueR = evaluateExp( tree.getSubTree(1) );
         if ( ! valueR.tag.equals(Value.INT_TAG) )  // runtime check
         {
            throw new EvalException("not a integer expression: "
                                     + tree.getSubTree(1) + "\n");
         }
         resultR = valueR.valueI;
      }

      if ( node.equals("+") )
      {
         if ( tree.degree() == 1 )
            result = resultL;
         else
         {
            result = resultL + resultR;

            for (int i = 2; i < tree.degree(); i++)
            {
               Value temp = evaluateExp( tree.getSubTree(i) );
               if ( ! temp.tag.equals(Value.INT_TAG) )  // runtime check
               {
                  throw new EvalException("not a integer expression: "
                                           + tree.getSubTree(i) + "\n");
               }
               result += temp.valueI;
            }
         }
      }
      else if ( node.equals("-") )
      {
         if ( 2 < tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }
         if ( tree.degree() == 1 )
            result = -resultL;
         else
            result = resultL - resultR;
      }
      else if ( node.equals("*") )
      {
         if ( 1 == tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }

         result = resultL * resultR;

         for (int i = 2; i < tree.degree(); i++)
         {
            Value temp = evaluateExp( tree.getSubTree(i) );
            if ( ! temp.tag.equals(Value.INT_TAG) )  // runtime check
            {
               throw new EvalException("not a integer expression: "
                                        + tree.getSubTree(i) + "\n");
            }
            result *= temp.valueI;
         }
      }
      else if ( node.equals("/") )
      {
         if ( 2 != tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }
         result = resultL / resultR;
      }
      else if ( node.equals("%") )
      {
         if ( 2 != tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }
         result = resultL % resultR;
      }
      else if ( node.equals("^") )
      {
         if ( 2 != tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }
         result = (int)Math.pow(resultL, resultR);
      }

      return new Value( result );
   }//evaluateAExp()
}//Evaluate
